use anyhow::{Context as _, Result};
use std::{
    borrow::Cow,
    fs,
    io::{Read as _, Write as _},
};

fn main() -> Result<()> {
    let app = App::try_new()?;
    app.run()
}

struct App {
    is_terminal: bool,
    paths: Vec<Cow<'static, str>>,
}

impl App {
    fn try_new() -> Result<Self> {
        use std::io::IsTerminal as _;
        let is_terminal = std::io::stdout().is_terminal();

        let mut args = std::env::args();
        let _exe_name = args.next().context("How do we not have an exe name?")?;
        let mut paths = vec![];
        for arg in args {
            paths.push(arg.into());
        }
        let paths = if paths.is_empty() {
            vec![".".into()]
        } else {
            paths
        };

        Ok(App { is_terminal, paths })
    }

    fn run(&self) -> Result<()> {
        for path in &self.paths {
            let metadata = fs::metadata(&**path)?;
            let mut lk = if metadata.is_dir() {
                Lk::try_new_ls(path)?
            } else {
                Lk::try_new_cat(path)?
            };
            while let Some(buf) = lk.poll()? {
                std::io::stdout().write_all(&buf)?;
            }
        }
        Ok(())
    }
}

enum Lk {
    Cat(Cat),
    Ls(Ls),
}

impl Lk {
    fn try_new_cat(path: &str) -> Result<Self> {
        Ok(Self::Cat(Cat::try_new(path)?))
    }

    fn try_new_ls(path: &str) -> Result<Self> {
        Ok(Self::Ls(Ls::try_new(path)?))
    }

    fn poll(&mut self) -> Result<Option<Cow<[u8]>>> {
        match self {
            Lk::Cat(x) => x.poll(),
            Lk::Ls(x) => x.poll(),
        }
    }
}

struct Cat {
    f: fs::File,
    buf: Vec<u8>,
}

impl Cat {
    fn try_new(path: &str) -> Result<Self> {
        let f = fs::File::open(path)?;
        let buf = vec![0u8; 4_096];
        Ok(Self { f, buf })
    }

    fn poll(&mut self) -> Result<Option<Cow<[u8]>>> {
        let bytes_read = self.f.read(&mut self.buf)?;
        let buf = &self.buf[0..bytes_read];
        if bytes_read == 0 {
            return Ok(None);
        }
        Ok(Some(buf.into()))
    }
}

struct Ls {
    d: fs::ReadDir,
}

impl Ls {
    fn try_new(path: &str) -> Result<Self> {
        let d = fs::read_dir(path)?;
        Ok(Self { d })
    }

    fn poll(&mut self) -> Result<Option<Cow<[u8]>>> {
        let Some(entry) = self.d.next() else {
            return Ok(None);
        };
        let entry = entry?;
        let path = entry.file_name();
        let path = path.to_str().context("file name is not valid Unicode")?;
        Ok(Some(format!("{path}\n").into_bytes().into()))
    }
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn test() -> Result<()> {
        let dir = tempfile::tempdir()?;

        Ok(())
    }
}
